53 const float *coeffs =
s->coeffs;
60 nb_samples =
FFMIN(
s->nb_samples,
s->n -
s->pts);
61 if (nb_samples <= 0) {
69 memcpy(
frame->data[0], coeffs +
s->pts, nb_samples *
sizeof(
float));
97static float *
make_lpf(
int num_taps,
float Fc,
float beta,
float rho,
98 float scale,
int dc_norm)
100 int i, m = num_taps - 1;
101 float *
h =
av_calloc(num_taps,
sizeof(*
h)), sum = 0;
109 for (
i = 0;
i <= m / 2;
i++) {
110 float z =
i - .5f * m, x = z *
M_PI, y = z * mult1;
111 h[
i] = x ?
sinf(Fc * x) / x : Fc;
119 for (
i = 0; dc_norm &&
i < num_taps;
i++)
128 static const float coefs[][4] = {
129 {-6.784957e-10, 1.02856e-05, 0.1087556, -0.8988365 + .001},
130 {-6.897885e-10, 1.027433e-05, 0.10876, -0.8994658 + .002},
131 {-1.000683e-09, 1.030092e-05, 0.1087677, -0.9007898 + .003},
132 {-3.654474e-10, 1.040631e-05, 0.1087085, -0.8977766 + .006},
133 {8.106988e-09, 6.983091e-06, 0.1091387, -0.9172048 + .015},
134 {9.519571e-09, 7.272678e-06, 0.1090068, -0.9140768 + .025},
135 {-5.626821e-09, 1.342186e-05, 0.1083999, -0.9065452 + .05},
136 {-9.965946e-08, 5.073548e-05, 0.1040967, -0.7672778 + .085},
137 {1.604808e-07, -5.856462e-05, 0.1185998, -1.34824 + .1},
138 {-1.511964e-07, 6.363034e-05, 0.1064627, -0.9876665 + .18},
140 float realm = logf(tr_bw / .0005f) / logf(2.f);
143 float b0 = ((c0[0] * att + c0[1]) * att + c0[2]) * att + c0[3];
144 float b1 = ((
c1[0] * att +
c1[1]) * att +
c1[2]) * att +
c1[3];
146 return b0 + (
b1 -
b0) * (realm - (
int)realm);
149 return .1102f * (att - 8.7f);
151 return .58417f *
powf(att - 20.96f, .4f) + .07886f * (att - 20.96f);
155static void kaiser_params(
float att,
float Fc,
float tr_bw,
float *beta,
int *num_taps)
157 *beta = *beta < 0.f ?
kaiser_beta(att, tr_bw * .5f / Fc): *beta;
158 att = att < 60.f ? (att - 7.95f) / (2.285f *
M_PI * 2.f) :
159 ((.0007528358f-1.577737e-05 * *beta) * *beta + 0.6248022f) * *beta + .06186902f;
160 *num_taps = !*num_taps ?
ceilf(att/tr_bw + 1) : *num_taps;
163static float *
lpf(
float Fn,
float Fc,
float tbw,
int *num_taps,
float att,
float *beta,
int round)
167 if ((Fc /= Fn) <= 0.f || Fc >= 1.f) {
172 att = att ? att : 120.f;
174 kaiser_params(att, Fc, (tbw ? tbw / Fn : .05f) * .5f, beta, num_taps);
178 *num_taps =
av_clip(n, 11, 32767);
180 *num_taps = 1 + 2 * (int)((
int)((*num_taps / 2) * Fc + .5f) / Fc + .5f);
183 return make_lpf(*num_taps |= 1, Fc, *beta, 0.f, 1.f, 0);
188 for (
int i = 0;
i < n;
i++)
194#define SQR(a) ((a) * (a))
207 float *pi_wraps, *work, phase1 = (phase > 50.f ? 100.f - phase : phase) / 50.f;
208 int i, work_len, begin, end, imp_peak = 0, peak = 0, ret;
209 float imp_sum = 0, peak_imp_sum = 0,
scale = 1.f;
210 float prev_angle2 = 0, cum_2pi = 0, prev_angle1 = 0, cum_1pi = 0;
212 for (
i = *
len, work_len = 2 * 2 * 8;
i > 1; work_len <<= 1, i >>= 1);
215 work =
av_calloc((work_len + 2) + (work_len / 2 + 1),
sizeof(
float));
218 pi_wraps = &work[work_len + 2];
220 memcpy(work, *
h, *
len *
sizeof(*work));
231 s->tx_fn(
s->tx, work, work,
sizeof(
float));
233 for (
i = 0;
i <= work_len;
i += 2) {
234 float angle =
atan2f(work[
i + 1], work[
i]);
235 float detect = 2 *
M_PI;
236 float delta = angle - prev_angle2;
243 delta = angle - prev_angle1;
247 pi_wraps[
i >> 1] = cum_1pi;
255 for (
i = 0;
i < work_len;
i++)
256 work[
i] *= 2.f / work_len;
258 for (
i = 1;
i < work_len / 2;
i++) {
260 work[
i + work_len / 2] = 0;
262 s->tx_fn(
s->tx, work, work,
sizeof(
float));
264 for (
i = 2;
i < work_len;
i += 2)
265 work[
i + 1] = phase1 *
i / work_len * pi_wraps[work_len >> 1] + (1 - phase1) * (work[
i + 1] + pi_wraps[
i >> 1]) - pi_wraps[
i >> 1];
267 work[0] =
exp(work[0]);
268 work[1] =
exp(work[1]);
269 for (
i = 2;
i < work_len;
i += 2) {
270 float x =
expf(work[
i]);
272 work[
i ] = x *
cosf(work[
i + 1]);
273 work[
i + 1] = x *
sinf(work[
i + 1]);
277 for (
i = 0;
i < work_len;
i++)
278 work[
i] *= 2.f / work_len;
281 for (
i = 0;
i <= (int) (pi_wraps[work_len >> 1] /
M_PI + .5f);
i++) {
283 if (
fabs(imp_sum) >
fabs(peak_imp_sum)) {
284 peak_imp_sum = imp_sum;
287 if (work[
i] > work[imp_peak])
291 while (peak &&
fabsf(work[peak - 1]) >
fabsf(work[peak]) && (work[peak - 1] * work[peak] > 0)) {
297 }
else if (phase1 == 1) {
298 begin = peak - *
len / 2;
300 begin = (.997f - (2 - phase1) * .22f) * *
len + .5f;
301 end = (.997f + (0 - phase1) * .22f) * *
len + .5f;
302 begin = peak - (begin & ~3);
303 end = peak + 1 + ((end + 3) & ~3);
312 for (
i = 0;
i < *
len;
i++) {
313 (*h)[
i] = work[(begin + (phase > 50.f ? *
len - 1 -
i :
i) + work_len) & (work_len - 1)];
315 *post_len = phase > 50 ? peak - begin : begin + *
len - (peak + 1);
318 work_len, pi_wraps[work_len >> 1] /
M_PI, peak, peak_imp_sum, imp_peak,
319 work[imp_peak], *
len, *post_len, 100.f - 100.f * *post_len / (*
len - 1));
331 float Fn =
s->sample_rate * .5f;
333 int i, n, post_peak, longer, ret;
338 if (
s->Fc0 >= Fn ||
s->Fc1 >= Fn) {
340 "filter frequency must be less than %d/2.\n",
s->sample_rate);
344 h[0] =
lpf(Fn,
s->Fc0,
s->tbw0, &
s->num_taps[0],
s->att, &
s->beta,
s->round);
345 h[1] =
lpf(Fn,
s->Fc1,
s->tbw1, &
s->num_taps[1],
s->att, &
s->beta,
s->round);
350 longer =
s->num_taps[1] >
s->num_taps[0];
351 n =
s->num_taps[longer];
354 for (
i = 0;
i <
s->num_taps[!longer];
i++)
355 h[longer][
i + (n -
s->num_taps[!longer]) / 2] +=
h[!longer][
i];
363 if (
s->phase != 50.f) {
379 for (
i = 0;
i < n;
i++)
380 s->coeffs[
i] =
h[longer][
i];
408#define AF AV_OPT_FLAG_AUDIO_PARAM|AV_OPT_FLAG_FILTERING_PARAM
409#define OFFSET(x) offsetof(SincContext, x)
414 {
"nb_samples",
"set the number of samples per requested frame",
OFFSET(nb_samples),
AV_OPT_TYPE_INT, {.i64=1024}, 1, INT_MAX,
AF },
415 {
"n",
"set the number of samples per requested frame",
OFFSET(nb_samples),
AV_OPT_TYPE_INT, {.i64=1024}, 1, INT_MAX,
AF },
422 {
"hptaps",
"set number of taps for high-pass filter",
OFFSET(num_taps[0]),
AV_OPT_TYPE_INT, {.i64=0}, 0, 32768,
AF },
423 {
"lptaps",
"set number of taps for low-pass filter",
OFFSET(num_taps[1]),
AV_OPT_TYPE_INT, {.i64=0}, 0, 32768,
AF },
431 .p.description =
NULL_IF_CONFIG_SMALL(
"Generate a sinc kaiser-windowed low-pass, high-pass, band-pass, or band-reject FIR coefficients."),
432 .p.priv_class = &sinc_class,
static enum AVSampleFormat sample_fmts[]
static int query_formats(const AVFilterContext *ctx, AVFilterFormatsConfig **cfg_in, AVFilterFormatsConfig **cfg_out)
const FFFilter ff_asrc_sinc
static AVFormatContext * ctx
static float * make_lpf(int num_taps, float Fc, float beta, float rho, float scale, int dc_norm)
static float safe_log(float x)
static const AVFilterPad sinc_outputs[]
static const AVOption sinc_options[]
static int fir_to_phase(AVFilterContext *ctx, float **h, int *len, int *post_len, float phase)
static int query_formats(const AVFilterContext *ctx, AVFilterFormatsConfig **cfg_in, AVFilterFormatsConfig **cfg_out)
static int activate(AVFilterContext *ctx)
static av_cold void uninit(AVFilterContext *ctx)
static float kaiser_beta(float att, float tr_bw)
static void invert(float *h, int n)
static void kaiser_params(float att, float Fc, float tr_bw, float *beta, int *num_taps)
static float * lpf(float Fn, float Fc, float tbw, int *num_taps, float att, float *beta, int round)
static int config_output(AVFilterLink *outlink)
AVFrame * ff_get_audio_buffer(AVFilterLink *link, int nb_samples)
Request an audio samples buffer with a specific set of permissions.
simple assert() macros that are a bit more flexible than ISO C assert().
#define av_assert0(cond)
assert() equivalent, that is always enabled.
int ff_filter_frame(AVFilterLink *link, AVFrame *frame)
Send a frame of data to the next filter.
int ff_outlink_frame_wanted(AVFilterLink *link)
Test if a frame is wanted on an output link.
Main libavfilter public API header.
#define i(width, name, range_min, range_max)
Public libavutil channel layout APIs header.
static __device__ float sqrtf(float a)
static __device__ float fabsf(float a)
static __device__ float fabs(float a)
static __device__ float ceilf(float a)
static const int sample_rates[]
static av_unused double sinc(double x, const double *params)
static av_cold void cleanup(FlashSV2Context *s)
@ AV_OPT_TYPE_INT
Underlying C type is int.
@ AV_OPT_TYPE_FLOAT
Underlying C type is float.
@ AV_OPT_TYPE_BOOL
Underlying C type is int.
#define AV_CHANNEL_LAYOUT_MONO
#define AVERROR_EOF
End of file.
#define AV_LOG_DEBUG
Stuff which is only useful for libav* developers.
#define AV_LOG_ERROR
Something went wrong and cannot losslessly be recovered.
double av_bessel_i0(double x)
0th order modified bessel function of the first kind.
AVSampleFormat
Audio sample formats.
static void scale(int *out, const int *in, const int w, const int h, const int shift)
static av_cold void uninit(AVBitStreamFilterContext *ctx)
static int activate(AVBitStreamFilterContext *ctx)
static int config_output(AVBitStreamFilterLink *outlink)
static int16_t mult(Float11 *f1, Float11 *f2)
#define FILTER_OUTPUTS(array)
static void ff_outlink_set_status(AVFilterLink *link, int status, int64_t pts)
Set the status field of a link from the source filter.
#define FFERROR_NOT_READY
Filters implementation helper functions and internal structures.
#define AVFILTER_DEFINE_CLASS(fname)
#define FILTER_QUERY_FUNC2(func)
#define NULL_IF_CONFIG_SMALL(x)
Return NULL if CONFIG_SMALL is true, otherwise the argument without modification.
static av_always_inline av_const double round(double x)
void * av_calloc(size_t nmemb, size_t size)
Memory handling functions.
static int adjust(int x, int size)
#define FF_ARRAY_ELEMS(a)
An AVChannelLayout holds information about the channel layout of audio data.
Describe the class of an AVClass context structure.
A link between two filters.
AVFilterContext * src
source filter
int sample_rate
samples per second
A filter pad used for either input or output.
This structure describes decoded (raw) audio or video data.
#define av_realloc_f(p, o, n)
av_cold void av_tx_uninit(AVTXContext **ctx)
Frees a context and sets *ctx to NULL, does nothing when *ctx == NULL.
av_cold int av_tx_init(AVTXContext **ctx, av_tx_fn *tx, enum AVTXType type, int inv, int len, const void *scale, uint64_t flags)
Initialize a transform context with the given configuration (i)MDCTs with an odd length are currently...
@ AV_TX_INPLACE
Allows for in-place transformations, where input == output.
@ AV_TX_FLOAT_RDFT
Real to complex and complex to real DFTs.
void(* av_tx_fn)(AVTXContext *s, void *out, void *in, ptrdiff_t stride)
Function pointer to a function to perform the transform.
static double b1(void *priv, double x, double y)
static double b0(void *priv, double x, double y)